"""Module for handling imports and exports with images."""
from __future__ import annotations
from typing import TYPE_CHECKING
import cv2
import numpy as np

from . import validate_extension
if TYPE_CHECKING:
    from .. import Signal2


def validate_filename(filename: str) -> None:
    """Validates the name of the file.

    Parameters
    ----------
    filename : str
        Name of the file to check.
    """
    valid_extensions = (
        "jpg",
        "jpeg",
        "png",
    )
    validate_extension(filename, valid_extensions)


def export_signal2(filename: str, signal2: Signal2) -> None:
    """Exports the given two dimensional signal."""
    validate_filename(filename)
    pass


def import_signal2(filename: str, channel="mean", norm=False, sf_ax0=1,
                   sf_ax1=1, sp_ax0=0, sp_ax1=0) -> Signal2:
    """Imports a two dimensional signal from a file.

    Parameters
    ----------
    filename : str
        File to read from.
    channel : {"mean", "r", "g", "b", 0, 1, 2}, optional
        How to handle images with multiple channels, by default "mean",
        which means it takes the mean of every channel.
    sf_ax0, sf_ax1 : float, optional
        Sampling frequency for each axis, by default 1.
    sp_ax0, sp_ax1 : float, optional
        Starting point for each axis, by default 0.

    Returns
    -------
    Signal2
        Read image.
    """
    validate_filename(filename)
    channels = {
        "mean": _import_s2_mean,
        "r": lambda vals, norm: _import_s2_channel(vals, 2, norm),
        "g": lambda vals, norm: _import_s2_channel(vals, 1, norm),
        "b": lambda vals, norm: _import_s2_channel(vals, 0, norm),
        0: lambda vals, norm: _import_s2_channel(vals, 0, norm),
        1: lambda vals, norm: _import_s2_channel(vals, 1, norm),
        2: lambda vals, norm: _import_s2_channel(vals, 2, norm),
    }
    signal = cv2.imread(filename)
    values = channels[channel](signal, norm)

    ax0_samp_period = 1 / sf_ax0
    ax1_samp_period = 1 / sf_ax1
    val_shape = np.shape(values)
    ax0 = np.arange(val_shape[0]) * ax0_samp_period - sp_ax0
    ax1 = np.arange(val_shape[1]) * ax1_samp_period - sp_ax1
    return ax0, ax1, values


def _import_s2_channel(values: np.ndarray, channel: int, norm=False):
    return values[:, :, channel] / 255 if norm else values[:, :, channel]


def _import_s2_mean(values: np.ndarray, norm=False):
    return np.mean(values, axis=2) / 255 if norm else np.mean(values, axis=2)



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from tqdm import tqdm

from ..config import C1_METHOD, C2_METHOD
if TYPE_CHECKING:
    from ..sgn import Signal1, Signal2

########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||||||| Signal1 ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


def c1(signal1: Signal1, method=C1_METHOD) -> Signal1:
    """Calculates the one dimensional Cosine transform of a given
    signal.

    In order to perform the calculation, a specific algorithm can be
    given as a parameter.

    Parameters
    ----------
    signal1 : Signal1
        One dimensional signal to calculate the Cosine transform.
    method : {"i", "ii", "iii", "iv", 1, 2, 3, 4}, optional
        Cosine transform to calculate, by default F1_METHOD

    Returns
    -------
    Signal1
        Signal representing the Cosine transform.
    """
    output = C1_METHODS[method](signal1)
    output.axis *= output.sampling_freq() / (2 * output.span())
    return output


def _calculate_i_1(signal1: Signal1) -> Signal1:
    output = signal1.clone()
    signal_len = len(output)
    new_values = np.zeros(signal_len)
    for k in tqdm(range(signal_len), "Calculating DCT-I"):
        temp = 0
        for n, x in enumerate(output.values[1:-1]):
            temp += x * np.cos(np.pi * n * k / (signal_len - 1))
        new_values[k] = (temp + 0.5 * (output.values[0] +
                         (-1 ** k) * output.values[-1]))
    output.values = new_values * np.sqrt(2 / (signal_len - 1))
    return output


def _calculate_ii_1(signal1: Signal1) -> Signal1:
    output = signal1.clone()
    signal_len = len(output)
    new_values = np.zeros(signal_len)
    for k in tqdm(range(signal_len), "Calculating DCT-II"):
        temp = 0
        for n, x in enumerate(output.values):
            temp += x * np.cos((np.pi * k / signal_len) * (n + 0.5))
        new_values[k] = temp
    output.values = new_values * np.sqrt(2 / signal_len)
    return output


def _calculate_iii_1(signal1: Signal1) -> Signal1:
    output = signal1.clone()
    signal_len = len(output)
    new_values = np.zeros(signal_len)
    for k in tqdm(range(signal_len), "Calculating DCT-III"):
        temp = 0
        for n, x in enumerate(output.values[1:]):
            temp += x * np.cos((np.pi * n / signal_len) * (k + 0.5))
        new_values[k] = temp + 0.5 * output.values[0]
    output.values = new_values * np.sqrt(2 / signal_len)
    return output


def _calculate_iv_1(signal1: Signal1) -> Signal1:
    output = signal1.clone()
    signal_len = len(output)
    new_values = np.zeros(signal_len)
    for k in tqdm(range(signal_len), "Calculating DCT-IV"):
        temp = 0
        for n, x in enumerate(output.values):
            temp += x * np.cos((np.pi / signal_len) * (n + 0.5) * (k + 0.5))
        new_values[k] = temp
    output.values = new_values * np.sqrt(2 / signal_len)
    return output


def _calculate_fft1(signal1: Signal1) -> Signal1:
    pass


C1_METHODS = {
    1: _calculate_i_1,
    2: _calculate_ii_1,
    3: _calculate_iii_1,
    4: _calculate_iv_1,
    "i": _calculate_i_1,
    "ii": _calculate_ii_1,
    "iii": _calculate_iii_1,
    "iv": _calculate_iv_1,
    "fft": _calculate_fft1,
}

########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||||||| Signal2 ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


def c2(signal2: Signal2, method=C2_METHOD) -> Signal2:
    """Calculates the two dimensional Cosine transform of a given
    signal.

    In order to perform the calculation, a specific algorithm can be
    given as a parameter.

    Parameters
    ----------
    signal2 : Signal2
        Two dimensional signal to calculate the Cosine transform.
    method : {"ii", "iv", 2, 4}, optional
        Cosine transform to calculate, by default F1_METHOD

    Returns
    -------
    Signal2
        Signal representing the Cosine transform.
    """
    output = C2_METHODS[method](signal2)
    output.ax0 *= output.ax0_sampling_freq() / output.ax0_span()
    output.ax1 *= output.ax1_sampling_freq() / output.ax1_span()
    return output


def _calculate_ii_2(signal2: Signal2) -> Signal2:
    output = signal2.clone()
    signal_shape = output.shape()
    N, M = signal_shape
    new_values = np.zeros(signal_shape)
    for k in tqdm(range(N), "Calculating 2D DCT-II"):
        for l in range(M):
            temp = 0
            for n in range(N):
                for m in range(M):
                    temp += output[n, m] * np.cos(np.pi * (m + 0.5)
                                                  * l / M) * np.cos(np.pi * (n + 0.5) * k / N)
            new_values[k, l] = temp
    output.values = new_values
    return output


def _calculate_iv_2(signal2: Signal2) -> Signal2:
    output = signal2.clone()
    signal_shape = output.shape()
    N, M = signal_shape
    new_values = np.zeros(signal_shape)
    for k in tqdm(range(N), "Calculating 2D DCT-IV"):
        for l in range(M):
            temp = 0
            for n in range(N):
                for m in range(M):
                    temp += output[n, m] * np.cos(np.pi * (2 * m + 1) * (2 * k + 1) / (
                        4 * N)) * np.cos(np.pi * (2 * n + 1) * (2 * l + 1) / (4 * M))
            new_values[k, l] = temp
    output.values = new_values
    return output


C2_METHODS = {
    2: _calculate_ii_2,
    4: _calculate_iv_2,
    "ii": _calculate_ii_2,
    "iv": _calculate_iv_2,
}



"""
=======
Chirper
=======

Chirper is a package that aims to provide different tools and
functionalities for analyzing and processing signals.

Subpackages
-----------
sgn
    Basic creation and manipulation of signals.
modulation
    Different methods for modulating and demodulating signals,
    particularly useful when using signals to transmit and receive
    information.
transforms
    Implementation of different integral transforms utilized in signal
    processing applications.
api
    An API for the GUI to request data in a nicely formatted way.
gui
    Subpackage that contains the code for the GUI, which allows for live
    signal visualization and manipulation.
"""

import os
from importlib.metadata import version

from .gui import main_gui


__all__ = ["sgn", "modulation", "transforms"]
__version__ = version("chirper-py")

BASE_DIRNAME = os.path.dirname(__file__)


def run():
    main_gui.main()



from ..sgn import Signal1
from ..sgn.defaults import COS, SIN
from ..config import AM_MODULATION, HERTZ, SSB_UPPER
from ..transforms import hilbert


def am_modulation(signal1: Signal1, *args, method=AM_MODULATION, hertz=HERTZ, **kwargs) -> Signal1:
    """Applies AM modulation to the given one dimensional signal.

    The currently available methods for modulation are:
     - DSBFC : Double-SideBand Full Carrier.
     - DSBSC : Double-SideBand Suppressed Carrier.
     - SSB : Single-SideBand.
     - USB : Upper-SideBand, an alias for calling SSB with `upper=True`.
     - LSB : Lower-SideBand, an alias for calling SSB with `upper=False`.

    Parameters
    ----------
    signal1 : Signal1
        One dimensional signal to modulate.
    carrier_freq : float
        Frequency of the carrier wave.
    carrier_amp : float
        Amplitude of the carrier wave. Only for DSBFC and DSBSC.
    method : {"dsbfc", "dsbsc", "ssb", "usb", "lsb"}, optional
        Method used for the modulation, by default AM_MODULATION.
    hertz : bool, optional
        Whether the frequency is given in Hertz, by default HERTZ.

    Returns
    -------
    Signal1
        Modulated one dimensional signal.
    """
    return AM_MODULATION_METHODS[method](signal1, *args, hertz, **kwargs)


def _dsbfc_modulation(signal1: Signal1, carrier_freq, carrier_amp, hertz=HERTZ) -> Signal1:
    copy = signal1.clone()
    axis = copy.axis
    carrier = COS(axis, carrier_freq, 1, hertz)
    return (carrier_amp + copy) * carrier


def _dsbsc_modulation(signal1: Signal1, carrier_freq, carrier_amp, hertz=HERTZ) -> Signal1:
    copy = signal1.clone()
    axis = copy.axis
    carrier = COS(axis, carrier_freq, carrier_amp, hertz)
    return carrier * copy


def _ssb_modulation(signal1: Signal1, carrier_freq, hertz=HERTZ, upper=SSB_UPPER) -> Signal1:
    x = signal1.clone()
    axis = x.axis
    x_h = hilbert.h1(x)
    return x * COS(axis, carrier_freq, 1, hertz) + ((-1) ** int(upper)) * x_h * SIN(axis, carrier_freq, 1, hertz)


def _usb_modulation(signal1: Signal1, carrier_freq, hertz=HERTZ) -> Signal1:
    return _ssb_modulation(signal1, carrier_freq, hertz, True)


def _lsb_modulation(signal1: Signal1, carrier_freq, hertz=HERTZ) -> Signal1:
    return _ssb_modulation(signal1, carrier_freq, hertz, False)


AM_MODULATION_METHODS = {
    "dsbfc": _dsbfc_modulation,
    "dsbsc": _dsbsc_modulation,
    "ssb": _ssb_modulation,
    "usb": _usb_modulation,
    "lsb": _lsb_modulation,
}



"""
===================
Integral transforms
===================

This subpackage gives access to different integral transforms used for
signal processing purposes. The ones currently available are:
- Fourier
- Inverse Fourier
- Hilbert
"""



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np

from .config import CONVOLUTION_METHOD, CROSS_CORRELATION_METHOD, KERNEL_OOB
from .exceptions import DimensionError
if TYPE_CHECKING:
    from .sgn import Signal1, Signal2

########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||||||| Signal1 ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


def convolution(s1_x: Signal1, s1_y: Signal1,
                method=CONVOLUTION_METHOD) -> Signal1:
    """Calculates the convolution of two one-dimensional signals.

    There are different methods to calculate the convolution of two
    signals. The ones currently implemented are:
     - "fft": Uses the property that convolution in the time domain
        translates into multiplication in the frequency domain.
        That way, one can use the FFT to calculate the Fourier
        transforms of both signals (which is done really quickly),
        multiplies the spectra and then applies the inverse.

     - "direct" : Uses the formula of convolution to calculate it via
        brute-force. Very inneficient, as it is O(N*M).

    Parameters
    ----------
    s1_x : Signal1
        First one-dimensional signal to convolute.
    s1_y : Signal1
        Second one-dimensional signal to convolute.
    method : {"fft", "direct"}, optional
        Method used for the convolution, by default CONVOLUTION_METHOD.

    Returns
    -------
    Signal1
        Convoluted signal.
    """
    conv_methods = {
        "fft": conv_fft,
        "direct": conv_direct,
    }
    return conv_methods[method](s1_x, s1_y)


def conv_fft(s1_x: Signal1, s1_y: Signal1) -> Signal1:
    """Convolutes using the FFT."""
    from .transforms import fourier, ifourier
    x_fourier = fourier.f1(s1_x)
    y_fourier = fourier.f1(s1_y)
    return ifourier.if1(x_fourier * y_fourier)


def conv_direct(s1_x: Signal1, s1_y: Signal1) -> Signal1:
    """Convolutes via brute-force."""
    x_copy = s1_x.clone()
    y_copy = s1_y.clone()
    if not np.array_equal(x_copy.axis, y_copy.axis):
        raise DimensionError("Dimensions of signals do not match.")
    vals = []
    for n, _ in enumerate(x_copy.values):
        sum = 0
        for m, _ in enumerate(y_copy.values):
            sum += x_copy.values[m] * y_copy.values[n - m]
        vals.append(sum)
    output = s1_x.clone()
    output.values = np.array(vals)
    return output


def cross_correlation(s1_x: Signal1, s1_y: Signal1,
                      method=CROSS_CORRELATION_METHOD) -> Signal1:
    """Calculates the cross correlation of two signals.

    Parameters
    ----------
    s1_x : Signal1
        First signal.
    s1_y : Signal1
        Second signal
    method : {"fft", "direct"}, optional
        Desired method to calculate the cross correlation, by default
        CROSS_CORRELATION_METHOD.

    Returns
    -------
    Signal1
        Cross correlated signal
    """
    cc_methods = {
        "fft": cc_fft,
        "direct": cc_direct,
    }
    return cc_methods[method](s1_x, s1_y)


def cc_direct(s1_x: Signal1, s1_y: Signal1) -> Signal1:
    x_copy = s1_x.clone()
    y_copy = s1_y.clone()
    if not np.array_equal(x_copy.axis, y_copy.axis):
        raise DimensionError("Dimensions of signals do not match.")
    vals = []
    for n, _ in enumerate(x_copy.values):
        sum = 0
        for m, _ in enumerate(y_copy.values):
            sum += np.conjugate(x_copy.values[m]) * \
                y_copy.values[(m + n) % len(x_copy)]
        vals.append(sum)
    output = s1_x.clone()
    output.values = np.array(vals)
    return output


def cc_fft(s1_x: Signal1, s1_y: Signal1) -> Signal1:
    from .transforms import fourier, ifourier
    x_copy = s1_x.clone()
    y_copy = s1_y.clone()
    x_fourier = fourier.f1(x_copy)
    y_fourier = fourier.f1(y_copy)
    return ifourier.if1(x_fourier.conjugate() * y_fourier)

########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||||||| Signal2 ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


def apply_kernel(signal2: Signal2, kernel: np.ndarray, flip=False,
                 oob=KERNEL_OOB) -> Signal2:
    """Applies a given kernel to the two dimensional signal.

    This operation can be often found in the literature as image
    convolution.

    Predefined kernels can be found within `chirper.kernel`, but a
    custom one can be given without a problem.

    Parameters
    ----------
    signal2 : Signal2
        Signal to apply the kernel to.
    kernel : np.ndarray
        Kernel used for the operation.
    flip : bool, optional
        Whether to flip the kernel or not, by default False.
    oob : {"zero"}, optional
        Determines how to handle the values out of the signal. For
        example, when `oob` is `"zero"` then, when trying to get
        the 6th element of a 5x5 signal you will just get a 0. By
        default KERNEL_OOB.

    Returns
    -------
    Signal2
        Signal with the kernel applied to it.
    """
    copy = signal2.clone()
    signal_shape = copy.shape()
    result = np.empty((signal_shape))
    ker_copy = kernel.copy().T if flip else kernel.copy()
    ker_shape = np.shape(ker_copy)
    for i in range(signal_shape[0]):
        for j in range(signal_shape[1]):
            sum = 0
            for row, col, ker_row, ker_col in _generate_indices(ker_shape, i, j):
                sum += _get(copy, row, col, oob) * ker_copy[ker_row, ker_col]
            result[i, j] = sum
    copy.values = result
    return copy


def _get(signal2: Signal2, row, col, oob=KERNEL_OOB):
    methods = {
        "zero": _get_zero,
    }
    copy = signal2.clone()
    row_total, col_total = copy.shape()
    if (0 <= row < row_total) and (0 <= col < col_total):
        return signal2[row, col]
    else:
        return methods[oob](signal2, row, col)


def _get_zero(signal2: Signal2, row, col):
    return 0


def _generate_indices(ker_shape, row, col):
    ker_rows, ker_cols = ker_shape
    center = ((ker_rows - 1) // 2, (ker_cols - 1) // 2)
    return [
        (row + i, col + j, center[0] + i, center[1] + j)
        for i in range(-center[0], center[0] + 1)
        for j in range(-center[1], center[1] + 1)
    ]

########################################################################################################################
# ||||||||||||||||||||||||||||||||||||||||||||||||| Others ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from tqdm import tqdm

from . import fourier
from .. import window
from ..sgn import Signal1, Signal2


def stft1(signal1: Signal1, time_interval=None,
          window_method="rectangular", samp_time=0.01,
          interp_method="linear", shift=True, scale=True,
          *args, **kwargs) -> Signal2:
    windows = {
        "rectangular": window.w_rectangular,
        "gaussian": window.w_gaussian,
    }
    copy = signal1.clone()
    w_signal = windows[window_method](samp_time, *args, **kwargs)

    if time_interval is None:
        time_interval = (signal1.axis[0], signal1.axis[-1])

    copy = copy.get(*time_interval)

    windowed = copy.apply_window(w_signal, 0.5 * samp_time, interp_method)
    w_fourier = fourier.f1(windowed)

    time_axis = np.arange(*time_interval, samp_time)
    freq_axis = w_fourier.axis
    values = np.zeros((1, len(freq_axis)))

    for t in tqdm(time_axis + 0.5 * samp_time, "Calculating STFT"):
        windowed = copy.apply_window(w_signal, t, interp_method)
        w_fourier = fourier.f1(windowed, shift=shift, scale=scale)
        values_pad, w_pad = _pad(values, w_fourier.values)

        # If we had to pad the previous values, we update the frequency axis
        if len(w_fourier.axis) > len(freq_axis):
            freq_axis = w_fourier.axis
        values = np.vstack((values_pad, w_pad))
    return Signal2(time_axis, freq_axis, values[1:, :])


def _pad(arr1, arr2):
    # Helper function to handle padding of matrices when stacking
    len1, len2 = arr1.shape, arr2.shape
    if len(len1) == 1:
        max_len = max(len1[0], len2[0])
        arr1_copy = np.pad(arr1, (0, max_len - len1[0]))
        arr2_copy = np.pad(arr2, (0, max_len - len2[0]))
        return arr1_copy, arr2_copy
    else:
        max_len = max(len1[1], len2[0])
        arr1_copy = np.pad(arr1, [(0, 0), (0, max_len - len1[1])])
        arr2_copy = np.pad(arr2, (0, max_len - len2[0]))
        return arr1_copy, arr2_copy



"""
=======
GUI API
=======

This subpackage contains the basic tools to request data from Chirper
in a way that facilitates showing data in a GUI.
"""
from __future__ import annotations
from typing import TYPE_CHECKING
from copy import deepcopy

from .request_handler import RequestHandler
from .input_source import InputSource
from .data_handler import DataHandler
from .data_process import DataProcess
from .chirp import Chirp, ChirpType, ChirpSource
if TYPE_CHECKING:
    from ..sgn import Signal1

########################################################################################################################
# ||||||||||||||||||||||||||||||||||||||||||||||| Chirp Types |||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


class ChirpTypeSpectrogram(ChirpType):
    def __str__(self) -> str:
        return "spectrogram"

    def get_processed(self, data_process: DataProcess, data, **kwargs):
        return data_process.process_spectrogram(data, **kwargs)

    def get_handled(self, data_handler: DataHandler, signal: Signal1, **kwargs):
        return data_handler.handle_spectrogram(signal, **kwargs)

    def fetch(self, input_source: InputSource, source: ChirpSource, **kwargs):
        return source.get_fetched(input_source, **kwargs)


class ChirpTypeStart(ChirpType):
    def __str__(self) -> str:
        return "start"

    def fetch(self, input_source: InputSource, source: ChirpSource, **kwargs):
        return source.start_stream(input_source, **kwargs)


class ChirpTypeStop(ChirpType):
    def __str__(self) -> str:
        return "stop"

    def fetch(self, input_source: InputSource, source: ChirpSource, **kwargs):
        return source.stop_stream(input_source, **kwargs)


class ChirpTypeClear(ChirpType):
    def __str__(self) -> str:
        return "clear"

    def get_processed(self, data_process: DataProcess, data, **kwargs):
        return 0

    def get_handled(self, data_handler: DataHandler, signal: Signal1, **kwargs):
        return data_handler.clear_data(signal, **kwargs)

    def fetch(self, input_source: InputSource, chirp_source: ChirpSource, **kwargs):
        return 0

########################################################################################################################
# ||||||||||||||||||||||||||||||||||||||||||||||| Chirp Sources |||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


class ChirpSourceMicrophone(ChirpSource):
    def __str__(self) -> str:
        return "microphone"

    def get_fetched(self, input_source: InputSource, **kwargs):
        return input_source.fetch_microphone(**kwargs)

    def start_stream(self, input_source: InputSource, **kwargs):
        return input_source.start_microphone(**kwargs)

    def stop_stream(self, input_source: InputSource, **kwargs):
        return input_source.stop_microphone(**kwargs)

########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||||| Base classes |||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


class GuiInterface:
    """Interface for the GUI to send instructions and receive data."""
    REQUEST_TYPES = {
        "spectrogram": ChirpTypeSpectrogram,
        "start": ChirpTypeStart,
        "stop": ChirpTypeStop,
        "clear": ChirpTypeClear,
    }
    REQUEST_SOURCES = {
        "microphone": ChirpSourceMicrophone,
    }

    def __init__(self) -> None:
        self.data_handler = DataHandler(self)
        self.input_source = InputSource(self)
        self.request_handler = RequestHandler(self)
        self.data_process = DataProcess(self)

    def make_request(self, request_data):
        request, kwargs = self.parse_request_data(request_data)
        result = self.request_handler.take_request(request, **kwargs)
        return result

    def parse_request_data(self, request_data: dict) -> Chirp:
        copy_data = deepcopy(request_data)
        data_type = copy_data.pop("request_type")
        data_source = copy_data.pop("source")

        # Make the request
        request_type = GuiInterface.REQUEST_TYPES[data_type]()
        source = GuiInterface.REQUEST_SOURCES[data_source]()
        return Chirp(request_type, source), copy_data



"""
Submodule that contains the functions used for importing and exporting
different signals to different files.

Currently, there is support for the following formats:
- csv
- json
- wav
"""
from multipledispatch import dispatch

from ...exceptions import InvalidFileExtension


@dispatch(str, str)
def validate_extension(filename: str, expected: str) -> None:
    """Validates the extension of the given filename.

    Parameters
    ----------
    filename : str
        Name of the file to check.
    expected : str
        Expected file extension.

    Raises
    ------
    InvalidFileExtension
        If the file extension is not valid.
    """
    extension = filename.split(".")[-1]
    if extension != expected:
        raise InvalidFileExtension(extension=extension, exp_extension=expected)


@dispatch(str, (list, tuple))
def validate_extension(filename: str, expected: list[str]) -> None:
    """Validates the extension of the given filename.

    Parameters
    ----------
    filename : str
        Name of the file to check.
    expected : list[str]
        List of expected file extensions.

    Raises
    ------
    InvalidFileExtension
        If the file extension is not valid.
    """
    extension = filename.split(".")[-1]
    if extension not in expected:
        raise InvalidFileExtension(extension=extension, exp_extension=expected)



import numpy as np


def ker_mean(size):
    """Creates a kernel that, when convoluted with a two dimensional
    signal, takes the mean over a given size. The `size` parameter
    determines how far the kernel will reach for the values.

    Parameters
    ----------
    size : int, odd
        Size of the kernel.

    Returns
    -------
    np.ndarray
        Kernel matrix.

    Raises
    ------
    ValueError
        If a `size` is given that isn't an odd integer.
    """
    if size % 2 != 1:
        raise ValueError("Size must be an odd integer.")
    return np.ones((size, size)) / (size ** 2)


def ker_edge(level=2):
    """Creates an edge detecting kernel.

    Parameters
    ----------
    level : int, optional
        Sensibility of the edge detection, by default 2.

    Returns
    -------
    np.ndarray
        Kernel matrix.
    """
    if level == 0:
        return np.array([
            [1, 0, -1],
            [0, 0, 0],
            [-1, 0, 1]
        ])
    elif level == 1:
        return np.array([
            [0, -1, 0],
            [-1, 4, -1],
            [0, -1, 0]
        ])
    elif level == 2:
        return np.array([
            [-1, -1, -1],
            [-1, 8, -1],
            [-1, -1, -1]
        ])


def ker_sharpen():
    """Creates a kernel that sharpens a two dimensional signal.

    Returns
    -------
    np.ndarray
        Kernel matrix.
    """
    return np.array([
        [0, -1, 0],
        [-1, 5, -1],
        [0, -1, 0],
    ])



"""Module for handling imports and exports with .wav files."""
from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from scipy.io import wavfile

from . import validate_extension
if TYPE_CHECKING:
    from .. import Signal1


def validate_filename(filename: str) -> None:
    """Validates the name of the file.

    Parameters
    ----------
    filename : str
        Name of the file to check.
    """
    validate_extension(filename, "wav")


def export_signal1(filename: str, signal1: Signal1, samp_rate=None) -> None:
    """Exports the given one dimensional signal to the .wav file."""
    validate_filename(filename)
    if samp_rate is None:
        samp_rate = int(signal1.sampling_freq())
    wavfile.write(filename, samp_rate, signal1.values.astype(np.float32))


def import_signal1(filename: str, amplification=1/200, channels="mean", *args, **kwargs) -> Signal1:
    """Imports a one dimensional signal from a .wav file."""
    channel_handler = {
        "mean": _mean,
        "get": _get,
    }
    validate_filename(filename)
    sf, values = wavfile.read(filename)
    axis = np.arange(values.shape[0]) / sf
    values = channel_handler[channels](values, *args, **kwargs)
    return axis, amplification * values


def _mean(values: np.ndarray):
    if len(values.shape) != 1:
        return values.mean(axis=1)
    else:
        return values


def _get(values: np.ndarray, channel=0):
    return values[:, :, channel]



from __future__ import annotations
from typing import TYPE_CHECKING
import sounddevice as sd
import logging

if TYPE_CHECKING:
    from . import GuiInterface
    from .chirp import Chirp


class InputSource:
    def __init__(self, api: GuiInterface) -> None:
        self.api = api
        self.source = None

    def fetch(self, request: Chirp, **kwargs):
        return request.request_type.fetch(self, request.source, **kwargs)

    def fetch_microphone(self, blocksize=4410, **kwargs):
        output = self.source.read(blocksize)
        self.api.blocksize = blocksize
        if output[1]:
            logging.info("Overflow detected")
        return output[0]

    def start_microphone(self, samplerate=44100, channels=1, **kwargs):
        self.source = sd.InputStream(samplerate, channels)
        self.api.samplerate = samplerate
        self.api.channels = channels
        self.source.start()

    def stop_microphone(self, **kwargs):
        self.source.stop()



import setuptools

with open("README.md", "r") as file:
    long_description = file.read()

setuptools.setup(
    name="chirper-py",
    version="1.1.1",
    author="Cristobal Allendes",
    author_email="callendes.molina@gmail.com",
    description="Package for signal analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/No-tengo-nombre/chirper",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.8",
)


from __future__ import annotations
from typing import TYPE_CHECKING
import logging

if TYPE_CHECKING:
    from . import GuiInterface
    from .chirp import Chirp


class RequestHandler:
    def __init__(self, api: GuiInterface) -> None:
        self.api = api

    def take_request(self, request: Chirp, **kwargs):
        data = self.send_to_source(request, **kwargs)

        # The source returns something if data is being fetched. If the
        # request is a control request, it is None
        if data is not None:
            logging.info(f"Received request | {request}")
            signal = self.send_to_process(data, request, **kwargs)
            return self.send_to_handler(signal, request, **kwargs)

    def send_to_source(self, request: Chirp, **kwargs):
        return self.api.input_source.fetch(request, **kwargs)

    def send_to_process(self, data, request: Chirp, **kwargs):
        return self.api.data_process.process(data, request, **kwargs)

    def send_to_handler(self, signal, request: Chirp, **kwargs):
        return self.api.data_handler.handle(signal, request, **kwargs)



class DimensionError(Exception):
    """Exception for innapropriate object dimensions."""

    def __init__(self, message, result, expected) -> None:
        self.message = message
        self.result = result
        self.expected = expected
        super().__init__(message, result, expected)

    def __str__(self) -> str:
        return f"{self.message} (expected {self.expected}, got {self.result})"


class InvalidModulation(Exception):
    """Exception for invalid calls to modulators."""
    pass


class InvalidFileExtension(Exception):
    """Exception for invalid file extensions."""

    def __init__(self, message="Invalid file extension", extension=None, exp_extension=None) -> None:
        self.message = message
        self.extension = extension
        self.exp_extension = exp_extension
        super().__init__(message, extension, exp_extension)

    def __str__(self) -> str:
        return f"{self.message} (expected {self.exp_extension}, was given {self.extension})"



import numpy as np

from ..sgn import Signal1
from ..config import HERTZ, FM_MODULATION


def fm_modulation(signal1: Signal1, *args, method=FM_MODULATION, hertz=HERTZ, **kwargs) -> Signal1:
    return FM_MODULATION_METHODS[method](signal1, *args, hertz, **kwargs)


def _trad_modulation(signal1: Signal1, carrier_freq, carrier_amp, const, hertz=HERTZ) -> Signal1:
    copy = signal1.clone()
    axis = copy.axis
    freq = 2 * np.pi * carrier_freq if hertz else carrier_freq
    values = carrier_amp * np.cos((freq + const * copy.values) * axis)
    return Signal1(axis, values)


FM_MODULATION_METHODS = {
    "trad": _trad_modulation,
}



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from tqdm import tqdm

from ..config import F1_METHOD, F2_METHOD
if TYPE_CHECKING:
    from ..sgn import Signal1, Signal2

########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||||||| Signal1 ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


def if1(signal1: Signal1, method=F1_METHOD, shift=True) -> Signal1:
    """Calculates the one dimensional Inverse Fourier transform of a
    given signal.

    In order to perform the calculation, a specific algorithm can be
    given as a parameter.

    Parameters
    ----------
    signal1 : Signal1
        One dimensional signal to calculate the Inverse Fourier
        transform.
    method : {"dft", "fft"}, optional
        Method used to calculate the transform, by default F1_METHOD
    shift : bool
        Whether to shift the frequencies or not before applying the
        inverse, by default True.

    Returns
    -------
    Signal1
        Signal representing the Inverse Fourier Transform.
    """
    if shift:
        output = freq_shift1(signal1)
    output.axis *= output.sampling_freq() / output.span()
    output = F1_METHODS[method](output)
    return output


def _calculate_dft1(signal1: Signal1) -> Signal1:
    """Calculates the Inverse Fourier Transform :math:`\\mathcal{F}^{-1}\\{X[k]\\} = x[n]` such that

        .. math::
        x[n] = \\sum_{k=0}^{N-1}X[k]e^{j2\\pi nk/N}

    Returns
    -------
    Signal representing the Inverse Fourier Transform.
    """
    output = signal1.clone()
    signal_len = len(output)
    new_values = np.zeros(signal_len, dtype=complex)
    for n in tqdm(range(signal_len), "Calculating Inverse DFT"):
        temp = 0 + 0j
        for k in range(signal_len):
            temp += output.values[k] * \
                np.exp(1j * (2 * n * k * np.pi / signal_len))
        new_values[n] = temp / signal_len
    output.values = new_values
    return output * signal_len


def _calculate_fft1(signal1: Signal1) -> Signal1:
    """Calculates the inverse FFT of a given signal.

    Parameters
    ----------
    signal1 : Signal1
        One dimensional signal to calculate the Inverse Fourier
        transform.

    Returns
    -------
    Signal1
        Signal representing the Inverse Fourier Transform.
    """
    output = signal1.clone()
    output.values = np.fft.ifft(output.values)
    return output


def freq_shift1(signal1: Signal1) -> Signal1:
    output = signal1.clone()
    signal_len = len(output)
    output.axis = output.axis + output.span() / 2
    output.values = np.array(
        [*output.values[signal_len // 2:], *output.values[:signal_len // 2]])
    return output


F1_METHODS = {
    "dft": _calculate_dft1,
    "fft": _calculate_fft1,
}

########################################################################################################################
# |||||||||||||||||||||||||||||||||||||||||||||||| Signal2 ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||| #
########################################################################################################################


def if2(signal2: Signal2, method=F2_METHOD, shift=True) -> Signal2:
    if shift:
        output = freq_shift2(signal2)
    output = F2_METHODS[method](output)
    # output.ax0 *= output.ax0_sampling_freq() / output.ax0_span()
    # output.ax1 *= output.ax1_sampling_freq() / output.ax1_span()
    output.ax0 = output.ax0 * output.ax0_sampling_freq() / output.ax0_span()
    output.ax1 = output.ax1 * output.ax1_sampling_freq() / output.ax1_span()
    return output


def _calculate_dft2(signal2: Signal2) -> Signal2:
    output = signal2.clone()
    ax0_len, ax1_len = output.shape()
    new_values = np.zeros((ax0_len, ax1_len), dtype=complex)
    for u in tqdm(range(ax0_len), "Calculating DFT"):
        for v in range(ax1_len):
            temp = 0 + 0j
            for n in range(ax0_len):
                for m in range(ax1_len):
                    temp += output.values[n, m] * \
                        np.exp(1j * 2 * np.pi *
                               (u * n / ax0_len + v * m / ax1_len))
            new_values[u, v] = temp
    output.values = new_values
    return output


def _calculate_fft2(signal2: Signal2) -> Signal2:
    output = signal2.clone()
    output.values = np.fft.ifft2(output.values)
    return output


def freq_shift2(signal2: Signal2) -> Signal2:
    """Shift the frequencies of the signal.

    Parameters
    ----------
    signal2 : Signal2
        Signal to shift.

    Returns
    -------
    Signal2
        Shifted signal
    """
    output = signal2.clone()
    ax0_len, ax1_len = output.shape()
    output.ax0 = output.ax0 + output.ax0_span() / 2
    output.ax1 = output.ax1 + output.ax1_span() / 2
    output.values = np.fft.ifftshift(output.values)
    return output


F2_METHODS = {
    "dft": _calculate_dft2,
    "fft": _calculate_fft2,
}
